package stdlib;

//- LJV.java --- Generate a graph of an object, using Graphviz
// The Lightweight Java Visualizer (LJV)
// https://www.cs.auckland.ac.nz/~j-hamer/

//- Author:     John Hamer <J.Hamer@cs.auckland.ac.nz>
//- Created:    Sat May 10 15:27:48 2003
//- Time-stamp: <2004-08-23 12:47:06 jham005>

//- Copyright (C) 2004  John Hamer, University of Auckland
//-
//-   This program is free software; you can redistribute it and/or
//-   modify it under the terms of the GNU General Public License
//-   as published by the Free Software Foundation; either version 2
//-   of the License, or (at your option) any later version.
//-
//-   This program is distributed in the hope that it will be useful,
//-   but WITHOUT ANY WARRANTY; without even the implied warranty of
//-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//-   GNU General Public License for more details.
//-
//-   You should have received a copy of the GNU General Public License along
//-   with this program; if not, write to the Free Software Foundation, Inc.,
//-   59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

//- $Id: LJV.java,v 1.1 2004/07/14 02:03:45 jham005 Exp $

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.lang.reflect.*;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

/* add methods:

   not introspective:
   drawArrayAsBinaryTree (heap)
   drawArrayAsForest     (union find, edgeTo.  roots=self or outside array)
   drawGraph
   drawDigraph
   drawWeightedGraph
   drawWeightedDigraph

   introspective (specify name of Node class as string... These could all be the same, perhaps, just picking out the Node class):
   drawList
   drawTree
   drawTrie
   drawTST
   drawChainingHashTable (maybe not necessary, if drawlist works)
   drawBTree (has Node and Entry)

   drawObjects()      use varargs
   drawNamedObjects() varargs alternating string, object, string, object...  Same object can appear more than once.
   drawObject()       for drawing an array as single object.
 */

@SuppressWarnings({ "rawtypes", "unchecked" })
public class YLJV {

	//- Set this to the path of the DOT executable from Graphviz.
	//- You need only set this if you choose an outputFormat different from "dot"
	private static String DOT_COMMAND = "/usr/local/bin/dot";
	//private static final String DOT_COMMAND = "c:/Program Files/ATT/Graphviz/bin/dot.exe";

	///** get the context in order set options */
	//public static Context options( ) { return defaultContext; }
	private static Context defaultContext = resetDefault( );
	/** Reset to default options  */
	public static Context resetDefault( ) {
		return resetVeryVerbose( )
				.ignoreField ("this$0")
				.treatPackageAsPrimitive( "java.lang" );
	}

	/** Show all of the structure in java.lang. */
	public static Context resetVeryVerbose( ) { return defaultContext = new Context( ); }

	/** Show most of the structure in java.lang. */
	public static Context resetVerbose( ) {
		return resetVeryVerbose( )
				.ignoreField( String.class, "hash" )
				.ignoreField( String.class, "hash32" );
	}

	public static class Context {
		private Context( ) { };

		/** Set the DOT output file format.  E.g., {@code ps} for PostScript, {@code png} for PNG, etc. */
		public Context setOutputFormat( String s ) { outputFormat = s; return this; }
		private String outputFormat = "png";

		/** The name of the output file is derived from {@code baseFileName} by appending successive integers. */
		public Context setBaseFileName( String s ) { baseFileName = s; return this; }
		private String baseFileName = "graph-";
		private static int nextGraphNumber = 0;
		private String nextFileName( ) {  return baseFileName + nextGraphNumber++ + "." + outputFormat; }
		private String keepDotFile = null;

		/**
		 * Set the DOT attributes for a class.  This allows you to change the
		 * appearance of certain nodes in the output, but requires that you
		 * know something about dot attributes.  Simple attributes are, e.g.,
		 * "color=red".
		 */
		public Context setClassAttribute( Class cz, String attrib ) { classAttributeMap.put( cz, attrib ); return this; }

		/** Set the DOT attributes for a specific field. */
		public Context setFieldAttribute( Field field, String attrib ) { fieldAttributeMap.put( field, attrib ); return this; }

		/** Set the DOT attributes for all fields with this name. */
		public Context setFieldAttribute( String field, String attrib ) { fieldAttributeMap.put( field, attrib ); return this; }

		/** Do not display this field. */
		public Context ignoreField( Field field ) { ignoreSet.add( field ); return this; }

		/** Do not display any fields with this name. */
		public Context ignoreField( String field ) { ignoreSet.add( field ); return this; }

		/** Do not display any fields with this name. */
		public Context ignoreField( Class cz, String field ) {
			Field[] fs = cz.getDeclaredFields( );
			for( int i = 0; i < fs.length; i++ )
				if (fs[i].getName( ).equals( field )) ignoreField( fs[i] );
			return this;
		}

		/** Do not display any fields from this class. */
		public Context ignoreFields( Class cz ) {
			Field[] fs = cz.getDeclaredFields( );
			for( int i = 0; i < fs.length; i++ )
				ignoreField( fs[i] );
			return this;
		}

		/** Do not display any fields with this type. */
		public Context ignoreClass( Class cz ) { ignoreSet.add( cz ); return this; }

		/** Do not display any fields that have a type from this package. */
		public Context ignorePackage( Package pk ) { ignoreSet.add( pk ); return this; }

		/** Treat objects of this class as primitives. */
		public Context treatClassAsPrimitive( Class cz ) { pretendPrimitiveSet.add( cz ); return this; }

		/** Treat objects from this package as primitives. */
		public Context treatPackageAsPrimitive( Package pk ) { pretendPrimitiveSet.add( pk ); return this; }
		public Context treatPackageAsPrimitive( String s ) { return treatPackageAsPrimitive( Package.getPackage( s ) ); }

		/** Hide private, protected and package-access fields. */
		public Context ignorePrivateFields( ) { ignorePrivateFields = true; return this; }
		private boolean ignorePrivateFields = false;

		/** Hide the field name in the label of an object. */
		//  TODO: allow this to be set on a per-class basis.
		public Context hideFieldNamesInLabels( ) { showFieldNamesInLabels = false; return this; }
		private boolean showFieldNamesInLabels = true;

		/** Show outer class names for nested classes. */
		public Context qualifyNestedClassNames( ) { qualifyNestedClassNames = true; return this; }
		private boolean qualifyNestedClassNames = false;

		/** Show package names for classes. */
		public Context showPackageNamesInClasses( ) { showPackageNamesInClasses = true; return this; }
		private boolean showPackageNamesInClasses = false;

		private Map classAttributeMap   = new HashMap( ); // Class,String
		private Map fieldAttributeMap   = new HashMap( ); // Field,String
		private Set pretendPrimitiveSet = new HashSet( ); // Class|Package
		private Set ignoreSet           = new HashSet( ); // Field|Class|Package
		private boolean canTreatAsPrimitive( Object obj ) {
			return obj == null || canTreatClassAsPrimitive( obj.getClass( ) );
		}
		private boolean canTreatClassAsPrimitive( Class cz ) {
			if( cz == null || cz.isPrimitive( ) ) return true;
			if( cz.isArray( ) ) return false;
			if( pretendPrimitiveSet.contains( cz ) || pretendPrimitiveSet.contains( cz.getPackage( ) ) ) return true;
			if( cz == Object.class ) return false;

			// Class sz = cz.getSuperclass( );
			// if (sz != null && sz != Object.class)
			//     return canTreatClassAsPrimitive( sz );
			// Class[] ifs = cz.getInterfaces( );
			// for( int i = 0; i < ifs.length; i++ )
			//     if( canTreatClassAsPrimitive( ifs[i] ) )
			//         return true;

			return false;
		}
		private boolean looksLikePrimitiveArray( Object obj ) {
			Class c = obj.getClass( );
			if( c.getComponentType( ).isPrimitive( ) )
				return true;

			for( int i = 0, len = Array.getLength( obj ); i < len; i++ )
				if( ! canTreatAsPrimitive( Array.get(obj, i) ) )
					return false;
			return true;
		}
		private boolean canIgnoreField( Field field ) {
			return Modifier.isStatic( field.getModifiers( ) )
					|| ignoreSet.contains( field )
					|| ignoreSet.contains( field.getName( ) )
					|| ignoreSet.contains( field.getType( ) )
					|| ignoreSet.contains( field.getType( ).getPackage( ) )
					;
		}
		protected String className( Object obj, boolean useToStringAsClassName ) {
			if( obj == null ) return "";
			Class c = obj.getClass( );
			if( useToStringAsClassName && redefinesToString( obj ) ) {
				return quote(obj.toString( ));
			} else {
				String name = c.getName( );
				if( ! showPackageNamesInClasses || c.getPackage( ) == YLJV.class.getPackage( ) ) {
					//- Strip away everything before the last .
					name = name.substring( name.lastIndexOf( '.' )+1 );

					if( ! qualifyNestedClassNames )
						name = name.substring( name.lastIndexOf( '$' )+1 );
				}
				return name;
			}
		}
	}

	//- Wrapper for objects that get visited.
	private static class VisitedObject {
		Object obj;
		VisitedObject( Object obj )           { this.obj = obj; }
		public boolean equals( Object other ) { return obj == ((VisitedObject)other).obj; }
		public int hashCode( )                { return System.identityHashCode( obj ); }
	}

	private static String dotName( Object obj ) {
		//- System.identityHashCode is a cheap way of generating a unique label for an object.
		return "n" + System.identityHashCode( obj );
	}

	private static boolean redefinesToString( Object obj ) {
		Method[] ms = obj.getClass( ).getMethods( );
		for( int i = 0; i < ms.length; i++ )
			if( ms[i].getName( ).equals( "toString" ) && ms[i].getDeclaringClass( ) != Object.class )
				return true;
		return false;
	}


	private static boolean fieldExistsAndIsPrimitive( Context ctx, Field field, Object obj ) {
		if( ! ctx.canIgnoreField( field ) ) {
			try {
				//- The order of these statements matters.  If field is not
				//- accessible, we want an IllegalAccessException to be raised
				//- (and caught).  It is not correct to return true if
				//- field.getType( ).isPrimitive( )
				Object val = field.get( obj );
				if( field.getType( ).isPrimitive( ) || ctx.canTreatAsPrimitive( val ) )
					//- Just calling ctx.canTreatAsPrimitive is not adequate --
					//- val will be wrapped as a Boolean or Character, etc. if we
					//- are dealing with a truly primitive type.
					return true;
			} catch( IllegalAccessException e ) { }
		}

		return false;
	}

	private static boolean hasPrimitiveFields( Context ctx, Field[] fs, Object obj ) {
		for( int i = 0; i < fs.length; i++ )
			if( fieldExistsAndIsPrimitive( ctx, fs[i], obj ) )
				return true;
		return false;
	}

	private static final String canAppearUnquotedInLabelChars = " $&*@#!-+()^%;[],;.=";
	private static boolean canAppearUnquotedInLabel( char c ) {
		return canAppearUnquotedInLabelChars.indexOf( c ) != -1
				|| Character.isLetter( c )
				|| Character.isDigit( c )
				;
	}

	private static final String quotable = "\"<>{}|";
	private static String quote( String s ) {
		StringBuffer sb = new StringBuffer( );
		for( int i = 0, n = s.length( ); i < n; i++ ) {
			char c = s.charAt(i);
			if( quotable.indexOf(c) != -1 )
				sb.append( '\\' ).append( c );
			else
				if( canAppearUnquotedInLabel( c ) )
					sb.append( c );
				else
					sb.append("\\\\0u" ).append( Integer.toHexString( c ) );
		}
		return sb.toString( );
	}

	protected static void processPrimitiveArray( Object obj, PrintWriter out ) {
		out.print( dotName( obj ) + "[shape=record, label=\"" );
		for( int i = 0, len = Array.getLength( obj ); i < len; i++ ) {
			if( i != 0 )
				out.print( "|" );
			out.print( quote(String.valueOf( Array.get(obj, i) )) );
		}
		out.println( "\"];" );
	}


	protected static void processObjectArray( Context ctx, Object obj, PrintWriter out, Set visited ) {
		out.print( dotName( obj ) + "[label=\"" );
		int len = Array.getLength( obj );
		for( int i = 0; i < len; i++ ) {
			if( i != 0 )
				out.print( "|" );
			out.print( "<f" + i + ">" );
		}
		out.println( "\",shape=record];" );
		for( int i = 0; i < len; i++ ) {
			Object ref = Array.get( obj, i );
			if( ref == null )
				continue;
			out.println( dotName( obj ) + ":f" + i + " -> " + dotName( ref )
			+ "[label=\"" + i + "\",fontsize=12];" );
			generateDotInternal( ctx, ref, out, visited );
		}
	}

	protected static void labelObjectWithSomePrimitiveFields( Context ctx, Object obj, Field[] fs, PrintWriter out ) {
		Object cabs  = ctx.classAttributeMap.get( obj.getClass( ) );
		out.print( dotName( obj ) + "[label=\"" + ctx.className( obj, false ) + "|{" );
		String sep = "";
		for( int i = 0; i < fs.length; i++ ) {
			Field field = fs[i];
			if( ! ctx.canIgnoreField( field ) )
				try {
					Object ref = field.get( obj );
					if( field.getType( ).isPrimitive( ) || ctx.canTreatAsPrimitive( ref ) ) {
						if( ctx.showFieldNamesInLabels )
							out.print( sep + field.getName( ) + ": " + quote(String.valueOf( ref )) );
						else
							out.print( sep + quote(String.valueOf( ref )) );
						sep = "|";
					}
				} catch( IllegalAccessException e ) { }
		}

		out.println( "}\"" + (cabs == null ? "" : "," + cabs) + ",shape=record];" );
	}

	protected static void labelObjectWithNoPrimitiveFields( Context ctx, Object obj, PrintWriter out ) {
		Object cabs  = ctx.classAttributeMap.get( obj.getClass( ) );
		out.println( dotName( obj )
				+ "[label=\"" + ctx.className( obj, true ) + "\""
				+ (cabs == null ? "" : "," + cabs)
				+ "];" );
	}

	protected static void processFields( Context ctx, Object obj, Field[] fs, PrintWriter out, Set visited ) {
		for( int i = 0; i < fs.length; i++ ) {
			Field field = fs[i];
			if( ! ctx.canIgnoreField( field ) ) {
				try {
					Object ref = field.get( obj );
					if( field.getType( ).isPrimitive( ) || ctx.canTreatAsPrimitive( ref ) )
						//- The field might be declared, say, Object, but the actual
						//- object may be, say, a String.
						continue;
					String name = field.getName( );
					Object fabs = ctx.fieldAttributeMap.get( field );
					if( fabs == null )
						fabs = ctx.fieldAttributeMap.get( name );
					out.println( dotName( obj ) + " -> " + dotName( ref )
					+ "[label=\"" + name + "\",fontsize=12"
					+ (fabs == null ? "" : "," + fabs)
					+ "];" );
					generateDotInternal( ctx, ref, out, visited );
				} catch( IllegalAccessException e ) { }
			}
		}
	}

	protected static void generateDotInternal( Context ctx, Object obj, PrintWriter out, Set visited ) {
		if( visited.add( new VisitedObject(obj) ) ) {
			if( obj == null ) {
				out.println( dotName( obj ) + "[label=\"null\"" + ", shape=plaintext];" );
			} else if (ctx.canTreatAsPrimitive( obj)) {
				out.println( dotName( obj ) + "[label=\"" + obj + "\"];" );
			} else {
				Class c = obj.getClass( );
				if( c.isArray( ) ) {
					if( ctx.looksLikePrimitiveArray( obj ) )
						processPrimitiveArray( obj, out );
					else
						processObjectArray( ctx, obj, out, visited );
				} else {
					Field[] fs = c.getDeclaredFields( );
					if( ! ctx.ignorePrivateFields )
						AccessibleObject.setAccessible( fs, true );

					if( hasPrimitiveFields( ctx, fs, obj ) )
						labelObjectWithSomePrimitiveFields( ctx, obj, fs, out );
					else
						labelObjectWithNoPrimitiveFields( ctx, obj, out );

					processFields( ctx, obj, fs, out, visited );

					//- If we cared, we would take the trouble to check which
					//- fields were accessible when we started, and carefully
					//- restore them.  Leaving them accessible does no real harm.
					// if( ! ctx.ignorePrivateFields )
					//   AccessibleObject.setAccessible( fs, false );
				}
			}
		}
	}

	/**
     Write a DOT digraph specification of the graph rooted at
     {@code obj} to {@code out}.
	 */
	public static void generateDOT( Context ctx, Object obj, PrintWriter out ) {
		out.println( "digraph Java {" );
		generateDotInternal( ctx, obj, out, new HashSet( ) );
		out.println( "}" );
	}

	public static void drawGraph( Context ctx, Object obj )       { drawGraph( ctx, obj, ctx.nextFileName( ) ); }
	public static void drawGraph( Object obj )                    { drawGraph( defaultContext, obj ); }
	public static void drawGraphToFile( Object obj, String file ) { drawGraph( defaultContext, obj, file ); }
	/** Create a graph of the object rooted at {@code obj}. */
	public static void drawGraph( Context ctx, Object obj, String file ) {
		try {
			if (ctx.outputFormat.equals ("dot")) {
				PrintWriter out;
				out = new PrintWriter( new FileWriter( new File( file ) ) );
				generateDOT( ctx, obj, out );
				out.close( );
			} else {
				File        dotfile = ctx.keepDotFile == null ? File.createTempFile( "LJV", "dot" ) : new File( ctx.keepDotFile );
				PrintWriter out     = new PrintWriter( new FileWriter( dotfile ) );
				try {
					generateDOT( ctx, obj, out );
					out.close( );
					ProcessBuilder pb = new ProcessBuilder( DOT_COMMAND,
							"-T" , ctx.outputFormat,
							dotfile.toString( ),
							"-o", file );
					pb.start ( ).waitFor( );
				} catch( InterruptedException e ) {
				} finally {
					if( ctx.keepDotFile == null )
						dotfile.delete( );
				}
			}
		} catch( IOException e ) {
			e.printStackTrace ();
			System.exit (-1);
		}
	}
}
